@@ -2,15 +2,17 @@ import copy
 import datetime
 import hashlib
 import importlib
+import importlib.util
 import inspect
 import functools
 import json
 import logging
-import os, stat, errno
+import os, stat, errno, sys
 import pickle
 import shutil
 from typing import Dict, List, Union, Callable
 from types import ModuleType
+from collections.abc import MutableMapping
 
 import networkx as nx
 import yaml
@@ -33,6 +35,29 @@ def handle_rmtree_error(delegate, path, exec):
 def rmtree(path):
     return shutil.rmtree(path, ignore_errors = False, onerror = handle_rmtree_error)
 
+def flatten(d, parent_key='', sep='.'):
+    items = []
+    for k, v in d.items():
+        new_key = str(parent_key) + sep + str(k) if parent_key else str(k)
+        if isinstance(v, MutableMapping):
+            items.extend(flatten(v, new_key, sep=sep).items())
+        else:
+            items.append((new_key, v))
+    return copy.deepcopy(dict(items))
+
+def unflatten(flatten, sep='.'):
+    out = {}
+    for key, value in flatten.items():
+        parts = key.split(sep)
+        d = out
+        for part in parts[:-1]:
+            if part.isdigit():
+                part = int(part)
+            if part not in d:
+                d[part] = {}
+            d = d[part]
+        d[parts[-1]] = value
+    return copy.deepcopy(out)
 
 class NoDefaultValue:
     pass
@@ -69,17 +94,40 @@ def get_stage_hash(descriptor):
     hash.update(source.encode("utf-8"))
     return hash.hexdigest()
 
+def synpp_import_module(name, package=None, externals={}):
+
+    absolute_name = importlib.util.resolve_name(name, package)
+    try:
+        return sys.modules[absolute_name]
+    except KeyError:
+        pass
+
+    if absolute_name in externals:
+        spec = importlib.util.spec_from_file_location(absolute_name, externals[absolute_name])
+    else:
+        spec = importlib.util.find_spec(absolute_name)
+        if spec is None:
+            msg = f'No module named {absolute_name!r}'
+            raise ModuleNotFoundError(msg, name=absolute_name)
+    module = importlib.util.module_from_spec(spec)
+    sys.modules[absolute_name] = module
+    spec.loader.exec_module(module)
+    return module
+
+
+def resolve_stage(descriptor, externals: dict = {}, aliases: dict = {}):
+    if descriptor in aliases:
+        descriptor = aliases[descriptor]
 
-def resolve_stage(descriptor):
     # Supported descriptors: module, class, @stage-decorated function or stage-looking object
     if isinstance(descriptor, str):
         # If a string, first try to get the actual object
         try:
-            descriptor = importlib.import_module(descriptor)
+            descriptor = synpp_import_module(descriptor, externals=externals)
         except ModuleNotFoundError:
             try:
                 parts = descriptor.split(".")
-                module = importlib.import_module(".".join(parts[:-1]))
+                module = synpp_import_module(".".join(parts[:-1]), externals=externals)
                 descriptor = getattr(module, parts[-1])
             except (ValueError, ModuleNotFoundError):
                 return None  # definitely not a stage
@@ -113,22 +161,38 @@ def get_config_path(name, config):
         return name
 
 def has_config_value(name, config):
-    for segment in name.split("."):
-        if segment in config:
-            config = config[segment]
-        else:
-            return False
+    splitted_req = name.split(".")
+    for key in flatten(config):
+        found = True
+        spltted_key = key.split(".")
+        for idx in range(len(splitted_req)):
+            if splitted_req[idx] != spltted_key[idx]:
+                found = False
+                break
+        if found:
+            return True
+    return False
 
-    return True
 
 def get_config_value(name, config):
-    for segment in name.split("."):
-        if segment in config:
-            config = config[segment]
-        else:
-            return None
+    flat = flatten(config)
+    if name in flat:
+        return flat[name]
 
-    return config
+    splitted_req = name.split(".")
+    values = []
+    keys = []
+    for key in flat:
+        found = True
+        splitted_key = key.split(".")
+        for idx in range(len(splitted_req)):
+            if splitted_req[idx] != splitted_key[idx]:
+                found = False
+                break
+        if found:
+            keys.append(".".join(splitted_key[len(splitted_req):]))
+            values.append(flat[key])
+    return unflatten(dict(zip(keys, values)))
 
 def configure_stage(instance, context, config):
     config_values = {}
@@ -183,9 +247,9 @@ class Context:
 
 
 class ConfigurationContext(Context):
-    def __init__(self, base_config, config_definitions=[]):
+    def __init__(self, base_config, config_definitions=[], externals={}, global_aliases={}):
         self.base_config = base_config
-        self.config_requested_stages = [resolve_stage(d).instance for d in config_definitions]
+        self.config_requested_stages = [resolve_stage(d, externals, global_aliases).instance for d in config_definitions]
 
         self.required_config = {}
 
@@ -194,6 +258,9 @@ class ConfigurationContext(Context):
 
         self.ephemeral_mask = []
 
+        self.externals = externals
+        self.global_aliases = global_aliases
+
     def config(self, option, default = NoDefaultValue()):
         if has_config_value(option, self.base_config):
             self.required_config[option] = get_config_value(option, self.base_config)
@@ -210,7 +277,7 @@ class ConfigurationContext(Context):
 
     def stage(self, descriptor, config = {}, alias = None, ephemeral = False):
         definition = {
-            "descriptor": descriptor, "config": config
+            "descriptor": descriptor, "config": flatten(config)
         }
 
         if not definition in self.required_stages:
@@ -222,7 +289,7 @@ class ConfigurationContext(Context):
 
     def stage_is_config_requested(self, descriptor):
         if self.config_requested_stages:
-            return resolve_stage(descriptor).instance in self.config_requested_stages
+            return resolve_stage(descriptor, self.externals, self.global_aliases).instance in self.config_requested_stages
         else:
             return False
 
@@ -260,10 +327,10 @@ class ExecuteContext(Context):
         self.progress_context = None
 
     def config(self, option):
-        if not option in self.required_config:
+        if not has_config_value(option, self.required_config):
             raise PipelineError("Config option %s is not requested" % option)
 
-        return self.required_config[option]
+        return get_config_value(option, self.required_config)
 
     def stage(self, name, config = {}):
         dependency = self._get_dependency({ "descriptor": name, "config": config })
@@ -299,7 +366,7 @@ class ExecuteContext(Context):
 
         return self.dependency_info[dependency][name]
 
-    def parallel(self, data = {}, processes = None, serialize = False):
+    def parallel(self, data = {}, processes = None, serialize = False, maxtasksperchild = None):
         config = self.required_config
 
         if processes is None and "processes" in self.pipeline_config:
@@ -311,7 +378,7 @@ class ExecuteContext(Context):
             # for profiling the code.
             return ParalelMockMasterContext(data, config, self.progress_context)
         else:
-            return ParallelMasterContext(data, config, processes, self.progress_context)
+            return ParallelMasterContext(data, config, processes, self.progress_context, maxtasksperchild)
 
     def progress(self, iterable = None, label = None, total = None, minimum_interval = 1.0):
         if minimum_interval is None and "progress_interval" in self.pipeline_config:
@@ -333,18 +400,20 @@ class ExecuteContext(Context):
         return self.dependencies[self.required_stages.index(definition)]
 
 
-def process_stages(definitions, global_config):
+def process_stages(definitions, global_config, externals={}, aliases={}):
     pending = copy.copy(definitions)
     stages = []
 
     for index, stage in enumerate(pending):
         stage["required-index"] = index
 
+    global_config = flatten(global_config)
+
     while len(pending) > 0:
         definition = pending.pop(0)
 
         # Resolve the underlying code of the stage
-        wrapper = resolve_stage(definition["descriptor"])
+        wrapper = resolve_stage(definition["descriptor"], externals, aliases)
         if wrapper is None:
             raise PipelineError(f"{definition['descriptor']} is not a supported object for pipeline stage definition!")
 
@@ -355,14 +424,14 @@ def process_stages(definitions, global_config):
             config.update(definition["config"])
 
         # Obtain configuration information through configuration context
-        context = ConfigurationContext(config, [d['descriptor'] for d in definitions])
+        context = ConfigurationContext(config, [d['descriptor'] for d in definitions], externals)
         wrapper.configure(context)
-
+        required_config = flatten(context.required_config)
         definition = copy.copy(definition)
         definition.update({
             "wrapper": wrapper,
-            "config": copy.copy(context.required_config),
-            "required_config": copy.copy(context.required_config),
+            "config": copy.copy(required_config),
+            "required_config": copy.copy(required_config),
             "required_stages": context.required_stages,
             "aliases": context.aliases
         })
@@ -381,7 +450,7 @@ def process_stages(definitions, global_config):
 
         # Process dependencies
         for position, upstream in enumerate(context.required_stages):
-            passed_parameters = list(upstream["config"].keys())
+            passed_parameters = set(flatten(upstream["config"]).keys())
 
             upstream_config = copy.copy(config)
             upstream_config.update(upstream["config"])
@@ -439,11 +508,9 @@ def process_stages(definitions, global_config):
 
             for upstream_index in stage["dependencies"]:
                 upstream = stages[upstream_index]
-
-                upstream_config_keys = upstream["config"].keys()
                 explicit_config_keys = upstream["downstream-passed-parameters"] if "downstream-passed-parameters" in upstream else set()
 
-                for key in upstream_config_keys - explicit_config_keys:
+                for key in upstream["config"].keys() - explicit_config_keys:
                     value = upstream["config"][key]
 
                     if key in passed_config_options:
@@ -505,7 +572,8 @@ def update_json(meta, working_directory):
 
 
 def run(definitions, config = {}, working_directory = None, flowchart_path = None, dryrun = False, verbose = False,
-        logger = logging.getLogger("synpp"), rerun_required=True, ensure_working_directory=False):
+        logger = logging.getLogger("synpp"), rerun_required=True, ensure_working_directory=False,
+        externals = {}, aliases = {}):
     # 0) Construct pipeline config
     pipeline_config = {}
     if "processes" in config: pipeline_config["processes"] = config["processes"]
@@ -521,7 +589,7 @@ def run(definitions, config = {}, working_directory = None, flowchart_path = Non
         working_directory = os.path.realpath(working_directory)
 
     # 1) Construct stage registry
-    registry = process_stages(definitions, config)
+    registry = process_stages(definitions, config, externals, aliases)
 
     required_hashes = [None] * len(definitions)
     for stage in registry.values():
@@ -822,15 +890,17 @@ def run_from_yaml(path):
 class Synpp:
     def __init__(self, config: dict, working_directory: str = None, logger: logging.Logger = logging.getLogger("synpp"),
                  definitions: List[Dict[str, Union[str, Callable, ModuleType]]] = None, flowchart_path: str = None,
-                 dryrun: bool = False):
+                 dryrun: bool = False, externals: Dict[str, str] = {}, aliases = {}):
         self.config = config
         self.working_directory = working_directory
         self.logger = logger
         self.definitions = definitions
         self.flowchart_path = flowchart_path
         self.dryrun = dryrun
+        self.externals = externals
+        self.aliases = aliases
 
-    def run_pipeline(self, definitions=None, rerun_required=False, dryrun=None, verbose=False, flowchart_path=None):
+    def run_pipeline(self, definitions=None, rerun_required=True, dryrun=None, verbose=False, flowchart_path=None):
         if definitions is None and self.definitions is None:
             raise PipelineError("A list of stage definitions must be available in object or provided explicitly.")
         elif definitions is None:
@@ -839,12 +909,13 @@ class Synpp:
             dryrun = self.dryrun
         return run(definitions, self.config, self.working_directory, flowchart_path=flowchart_path,
                    dryrun=dryrun, verbose=verbose, logger=self.logger, rerun_required=rerun_required,
-                   ensure_working_directory=True)
+                   ensure_working_directory=True, externals=self.externals, aliases=self.aliases)
 
     def run_single(self, descriptor, config={}, rerun_if_cached=False, dryrun=False, verbose=False):
         return run([{'descriptor': descriptor, 'config': config}], self.config, self.working_directory,
                    dryrun=dryrun, verbose=verbose, logger=self.logger, rerun_required=rerun_if_cached,
-                   flowchart_path=self.flowchart_path, ensure_working_directory=True)[0]
+                   flowchart_path=self.flowchart_path, ensure_working_directory=True, externals=self.externals,
+                   aliases=self.aliases)[0]
 
     @staticmethod
     def build_from_yml(config_path):
@@ -869,9 +940,11 @@ class Synpp:
         working_directory = settings["working_directory"] if "working_directory" in settings else None
         flowchart_path = settings["flowchart_path"] if "flowchart_path" in settings else None
         dryrun = settings["dryrun"] if "dryrun" in settings else False
+        externals = settings["externals"] if "externals" in settings else {}
+        aliases = settings["aliases"] if "aliases" in settings else {}
 
         return Synpp(config=config, working_directory=working_directory, definitions=definitions,
-                     flowchart_path=flowchart_path, dryrun=dryrun)
+                     flowchart_path=flowchart_path, dryrun=dryrun, externals=externals, aliases=aliases)
 
 
 def stage(function=None, *args, **kwargs):
