@@ -66,9 +66,9 @@ class ProgressTracker:
 class ProgressServer(threading.Thread):
     def __init__(self, tracker):
         threading.Thread.__init__(self)
-        context = zmq.Context()
+        self.context = zmq.Context()
 
-        self.socket = context.socket(zmq.PULL)
+        self.socket = self.context.socket(zmq.PULL)
         self.port = self.socket.bind_to_random_port("tcp://*")
         self.running = True
 
@@ -90,17 +90,25 @@ class ProgressServer(threading.Thread):
 
     def stop(self):
         self.running = False
+        time.sleep(0.001)
+
+        self.socket.close()
+        self.context.term()
 
 class ProgressClient:
     def __init__(self, port):
-        context = zmq.Context()
+        self.context = zmq.Context()
 
-        self.socket = context.socket(zmq.PUSH)
+        self.socket = self.context.socket(zmq.PUSH)
         self.socket.connect("tcp://localhost:%d" % port)
 
     def update(self, amount = 1):
         self.socket.send_string(json.dumps(dict(amount = amount)))
 
+    def close(self):
+        self.socket.close()
+        self.context.term()
+
 class ProgressContext:
     def __init__(self, iterable = None, total = None, label = None, logger = logging.getLogger("synpp"), minimum_interval = 0):
         self.tracker = ProgressTracker(total, label, logger, minimum_interval)
